# coding: latin-1

"""Dokumentation der fachlischen Konfigurationsvariablen

Dict mit Variablenname als Schlssel.

Wert ist ein tuple mit vier Feldern:

- fachlich: boolean, falls True ist die Variable vom Anwendungsadministrator editierbar.
- valid_pattern: regex oder None zur Validierung.
- beschreibung: Kurzbeschreibung (mglichst eine Zeile)
- doku: Weitere Dokumentation
"""

import re

def re_wohnt_nicht_ausserhalb():
    """
    Ist etwas komplizierter, daher ausgelagert,
    >>>
    >>> bool(re.match(re_wohnt_nicht_ausserhalb(), "ort:Berlin"))
    True
    >>> bool(re.match(re_wohnt_nicht_ausserhalb(), "ort:Berlin;bezirk:Treptow"))
    True
    >>> bool(re.match(re_wohnt_nicht_ausserhalb(), "ort :Berlin;Brandenburg"))
    True
    >>> bool(re.match(re_wohnt_nicht_ausserhalb(), "plz:Berlin"))
    False
    >>> bool(re.match(re_wohnt_nicht_ausserhalb(), "plraum:Berlin"))
    True
    >>> bool(re.match(re_wohnt_nicht_ausserhalb(), " plz:12345; 65432;ortsteil:Grobeeren"))
    True
    >>> bool(re.match(re_wohnt_nicht_ausserhalb(), " plz:1234; 65432;ortsteil:Grobeeren"))
    False
    >>> bool(re.match(re_wohnt_nicht_ausserhalb(), "plz:12345; 65432;\\nortsteil:Grobeeren;\\nwe123 ; samtgemeinde:a;b;c"))
    True
    """
    res = re.compile(r"""
         ^\s*$|                             # kann leer sein
           ^(
              (\s*(bezirk|ort|ortsteil|samtgemeinde|plraum)\s*:\s*[^;:]+\s*(;\s*[^;:]+\s*)*)
              |
              (\s*plz\s*:\s*\d{5,5}\s*(;\s*\d{5,5}\s*)*)
            )
           (;
              (\s*(bezirk|ort|ortsteil|samtgemeinde|plraum)\s*:\s*[^;:]+\s*(;\s*[^;:]+\s*)*)
              |
              (\s*plz\s*:\s*\d{5,5}\s*(;\s*\d{5,5}\s*)*)
           )*
          $""", re.VERBOSE)
    return res

zda_doku = """
Die drei Konfigurationsvariablen <code>wiederaufnahmefrist, zda_rueckgaengig_frist</code> und 
<code>zda_rueckgaengig_bis</code> legen fest, zu welchen Zeiten die Buttons "ZdA rckgngig"
und "Wiederaufnahme" auf der Klientenkarte in der Rubrik "Stand" zur Verfgung stehen.
<br><br>
<code>wiederaufnahmefrist</code>: Anzahl der Monate nach einem Fallabschluss, 
nach deren Ablauf der Button "Wiederaufnahme" erscheint, 
der die Fortfhrung der Beratung mit einer neuen Fallnummer einleitet.
Der Wert muss zwischen 0 und 99 liegen.
<br><br>
<code>zda_rueckgaengig_frist</code>: Anzahl der Monate nach einem Fallabschluss, 
innerhalb derer der Button "ZdA rckgngig" zur Verfgung steht, 
der die Fortfhrung der Beratung mit derselben Fallnummer ermglicht.
Der Wert muss zwischen -99 und 99 liegen. Falls der Wert negativ ist, 
wird vom Programm der Wert von <code>wiederaufnahmefrist</code> genommen.
<br><br>
<code>zda_rueckgaengig_bis</code>: Tag und Monat im Jahr, 
bis zu dem der Button "ZdA rckgngig" erscheinen darf. Der Wert muss entweder leer sein, 
oder das Format Tag.Monat annehmen. Ist der Wert z.B. 15.2, darf der Button "ZdA rckgngig" 
nur bis zum 15. Februar eines Jahres erscheinen, wenn der Fall im Vorjahr abgeschlossen wurde.
Damit kann verhindert werden, dass Flle mit der gleichen Fallnummer fortgefhrt werden,
wenn bereits fr diesen Fall eine Bundesstatistik fr einen abgeschlossenen Fall abgegeben
wurde. Falls ein Wert fr diese Variable angegeben wird, hat dieser Vorrang vor dem Wert der Variablen
<code>zda_rueckgaengig_frist</code>.
<br><br>
Es ergeben sich mit diesen drei Variablen also zum Beispiel folgende Mglichkeiten:
<ul>
<li><code>wiederaufnahmefrist</code>: 6<br>
<code>zda_rueckgaengig_frist</code>: -1 (Standardwert)<br>
<code>zda_rueckgaengig_bis</code>:&nbsp;&nbsp;  (leer, Standardwert)<br>
Wie in frheren Versionen erscheint 6 Monate der Button "ZdA rckgngig" 
und danach der Button "Wiederaufnahme". Ist der Wert 0, 
erscheint der Button "ZdA rckgngig" zu keinem Zeitpunkt.
</li>
<li><code>wiederaufnahmefrist</code>: 3<br>
<code>zda_rueckgaengig_frist</code>: 6<br>
<code>zda_rueckgaengig_bis</code>:&nbsp;&nbsp;  (leer, Standardwert)<br>
Der Button "ZdA rckgngig" erscheint bis 6 Monate nach Fallabschluss.
Ab 3 Monate nach Fallabschluss erscheint der Button "Wiederaufnahme". Es 
ergeben sich also drei Monate, in denen beide Buttons erscheinen, der Anwender also 
die Wahl hat zwischen "ZdA rckgngig" und "Wiederaufnahme". Es ist unzulssig, dass
der Wert fr <code>zda_rueckgaengig_frist</code> kleiner ist als der fr
<code>wiederaufnahmefrist</code>, da es dann eine Periode gbe, in der kein Button
erschiene.
<li><code>wiederaufnahmefrist</code>: 3<br>
<code>zda_rueckgaengig_frist</code>: 6<br>
<code>zda_rueckgaengig_bis</code>: 15.2<br>
Wie im vorigen Fall, aber der Button "ZdA rckgngig" erscheint nur
bis zum 15.2 des Jahres, falls der Fall im Vorjahr abgeschlossen wurde, auch
dann, wenn die durch <code>zda_rueckgaengig_frist</code> definierte Anzahl von Monaten
noch nicht erreicht sind.
</li>
</ul>
"""

param_doks = {
    'wiederaufnahmefrist':
        (True, 
         r"^\d{1,2}$",
         'Anzahl der Monate nach einem Fallabschluss, nach der der Button "Wiederaufnahme" erscheint,  '
         "der die Fortfhrung der Beratung mit einer neuen Fallnummer einleitet. ",
         zda_doku,
         ),
    
    'zda_rueckgaengig_frist':
        (True, 
         r"^[-]?\d{1,2}$",
         'Anzahl der Monate nach einem Fallabschluss, innerhalb derer der Button "ZdA rckgngig" zur Verfgung steht, '
         "der die Fortfhrung der Beratung mit derselben Fallnummer ermglicht. ",
         zda_doku,
         ),

    'zda_rueckgaengig_bis':
        (True,
         r"^$|^([1-9]|[1-2][0-9]|3[01])\.([1-9]|1[0-2])\.?$",
         'Tag und Monat im Jahr, bis zu dem der Button "ZdA rckgngig" erscheinen darf. ',
         zda_doku,
         ),
    
    'loeschfrist':
        (True,
         r"^\d{1,3}$",
         "Lschfrist in Monaten fr Aufbewahrungskategorie 'Beratung'.",
         "Akten, die lter sind (letzter Fallabschluss), knnen vom Administrator gelscht werden.",
         ),                                               

    'minimale_loeschfrist':
        (True,
         r"^\d{1,3}$",
         "Minimale Lschfrist, die als abweichende Lschfrist individuell fr einen Klienten eingetragen werden kann.",
         "",
         ),                                               
    
    'loeschfrist_therapie':
        (True,
         r"^\d{1,3}$",
         "Lschfrist in Monaten fr Aufbewahrungskategorie 'Therapie'.",
         "Akten, die lter sind (letzter Fallabschluss), knnen vom Administrator gelscht werden.",
         ),                                               

    'loeschfrist_missbrauch':
        (True,
         r"^\d{1,3}$",
         "Lschfrist in Monaten fr Aufbewahrungskategorie 'Missbrauch'.",
         "Akten, die lter sind (letzter Fallabschluss), knnen vom Administrator gelscht werden.",
         ),                                               

    'session_time':
        (True,
         r"^[123456789]\d{0,2}$",
         "Lnge einer Session in Minuten. ",
         "Nach dieser Zeit wird ein Benutzer"
         " automatisch abgemeldet, wenn er keine Eingaben mehr macht."
         ),
    
    'admin_name':
        (True,
         None,
         "Name des Anwendungsadministrators.",
         "Wird in Fehlermeldungen genannt.",
         ),
    
    'admin_tel':
        (True,
         None,
         "Telefonnummer des Anwendungsadministrators.",
         "Wird in Fehlermeldungen genannt.",
         ),
    
    'admin_email':
        (True,
         None,
         "Email-Adresse des Anwendungsadministrators.",
         "Wird in Fehlermeldungen genannt.",
         ),
    
    'instance_title':
        (True,
         None,
         """Langname fr die Instanz.""",
         """Erscheint bei der
Anmeldung und in der Statuszeile.""",
         ),
    
    'beratungskontakte':
        (True,
         None,
         """Falls <code>true</code> knnen in der
Klientenakte einzelne Beratungskontakte eingetragen werden.""",
         """Die Summen werden in die Bundesstatistik und in die 
Fachstatistik als Vorgabe bernommen.""",
         ),
    
    'beratungskontakte_bs':
        (True,
         None,
         """Falls <code>true</code> treten Anpassungen speziell
fr Braunschweig in Kraft.""",
         """Die Variable <code>beratungskontakte</code>
muss ebenfalls auf <code>true</code> stehen."""
         ),
    
    'fallunabhaengige_aktivitaeten_bs':
        (True,
         None,
         """Falls <code>true</code> knnen
fallunabhngige Aktivitten eingetragen und ausgewertet
werden. Speziell fr Braunschweig.""",
         ),
    
    'strassenkatalog':
        (True,
         None,
         """Falls diese Variable nicht leer ist, wird bei der Adresseingabe der Straenkatalog
verwendet.""",
         """Der eigentliche Wert der Variable spielt nur bei der Erstinstallation
der Instanz eine Rolle.<br>Eine Instanz ohne Straenkatalog kann auch spter noch fr die Verwendung 
mit Straenkatalog eingerichtet werden, indem der Straenkatalog ber 	
"Administratorhauptmen -- Straenkatalog importieren" importiert wird und der Variable 'strassenkatalog' ein
beliebiger Wert zugewiesen wird, beispielsweise der Ortsname.""",
         ),
    
    'strassenkatalog_vollstaendig':
        (True,
         None,
         """Falls <code>true</code> wird
angenommen, dass der Straenkatalog Eintrge fr jede einzelne
Hausnummer enthlt.""",
         """Ansonsten wird angenommen, dass lediglich Eintrge fr ganze Straen
bzw. fr bestimmte Intervalle von Hausnummern vorhanden sind.""",
         ),
    
    'strassensuche':
        (True,
         r"^(\s*(ort|bezirk|ortsteil|samtgemeinde)\s*)*$",
         """Bestimmt die Anzeige von Zusatzinformationen aus dem
Straenkatalog bei der Darstellung und Eingabe von Adressen.""",
         """Zulssige Werte sind: <code>ort bezirk ortsteil samtgemeinde</code>.<br><br>
Beispiele:
<ul>
<li>In Berlin gibt es nur einen Ort im Straenkatalog, nmlich 'Berlin'. Das Feld <code>ort</code> ist daher 
nicht ntig. Ortsteil und Samtgemeinde haben im Straenkatalog keine Werte, Bezirk allerdings schon. Der sinnvolle
Wert fr Berlin ist also
<pre>
    bezirk
</pre>
</li>
<li>In Gifhorn gibt es mehrere Orte im Straenkatalog. Sowohl Ortsteile und Samtgemeinden spielen eine Rolle und sind
im Straenkatalog erfasst, Bezirke jedoch nicht. Der sinnvolle
Wert hier ist also
<pre>
    ort ortsteil samtgemeinde
</pre>
</li>
<li>Wenn es nur einen Ort gibt und weder Bezirke, Ortsteile oder Samtgemeinden im Straenkatalog erfasst sind,
sollte die Variable leer bleiben.
</ul>
<br><br>Irrelevant, wenn kein Straenkatalog verwendet wird.""",
         ),
    
    'anmeldungsdaten_obligatorisch':
        (True,
         None,
         """Falls <code>true</code>
werden die Anmeldungsdaten obligatorisch im Aufnahmeformular erhoben anstatt optional 
in einer getrennten Maske.""",
         """Die entsprechenden Pflichfelder sind dabei
auszufllen. Eine sptere Bearbeitung der Anmeldungsdaten erfolgt
wie bisher in einer eigenstndigen Maske (analog der spteren
Bearbeitung einer Leistung).""",
         ),
    
    'sql_abfrage':
        (True,
         None,
         """Falls <code>true</code>
steht die Maske zum Absetzen von SQL-Abfragen vom
Administrationsmen aus zur Verfgung.""",
         '',
         ),
    
    'sql_sicherung':
        (True,
         None,
         """Falls <code>true</code>
steht ein Button im Administrationsmen zur Verfgung, der eine SQL-Datenbanksicherung auslst.""",
         '',
         ),
    
    'warnung_bei_fachstatistik_akte_diskrepanz':
        (True,
         None,
         """Falls <code>true</code>
wird beim Speichern der Fachstatistik auf eine Diskrepanz 
zwischen Akte und Fachstatistik hingewiesen, falls vorhanden.""",
         """Wenn man bei dem
dann zur Verfgung stehendem Item "Diskrepanz zur Akte zulassen" ein Hkchen
setzt, kann man trotzdem speichern.<br>
Betroffen sind nur die Items 'Familienstatus',
'Empfohlen von' und 'Erbrachte Leistungen in der Fachstatistik', deren
Werte im Normalfall aus der Akte vorbelegt werden.""",
         ),
    
    'fachstatistik_akte_diskrepanz_nicht_zulassen':
        (True,
         None,
         """Falls <code>true</code>
werden Diskrepanzen zwischen Fachstatistik und Akte nicht zugelassen.""",
         """Die entsprechenden Items in der Fachstatistik zeigen die Werte
aus der Akte und sind nicht modifizierbar. Eine Diskrepanz kann
so gar nicht entstehen. Allerdings muss bei nderungen in der
Akte die Fachstatistik neu abgespeichert werden, damit die
nderungen bernommen und fr die Statistik wirksam werden.<br>
Betroffen sind nur die Items 'Familienstatus',
'Empfohlen von' und 'Erbrachte Leistungen in der Fachstatistik', deren
Werte im Normalfall aus der Akte vorbelegt werden.""",
         ),

    'wohnt_nicht_ausserhalb':
        (True,
         re_wohnt_nicht_ausserhalb(),
         """Definiert die Teilmenge der Adressen, die im Sinne der Bundesstatistik in die Zustndigkeit
der Beratungsstelle fallen.""",
         """Aufgrund der hier angegebenen Kriterien wird entschieden, ob der Klient
in der Bundesstatistik als ausserhalb des Kreises der Beratungsstelle wohnend angegeben wird.
Als mgliche Kriterien kommen in Frage: <code>plz,ort,ortsteil,samtgemeinde,bezirk,plraum</code>.
Die Kriterien <code>ortsteil,samtgemeinde,bezirk</code> sind nur sinnvoll, wenn ein Straenkatalog
verwendet wird und dieser auch entsprechende Angaben enthlt.
<br><br>
Die Kriterien werden wie folgt in der angegebenen Reihenfolge abgearbeitet:
<ul>
<li>Wenn die Klientenadresse keinen Wert fr das Kriterium aufweist, wird zum nchsten
Kriterium bergegangen.
</li>
<li>Falls die Klientenadresse einen Wert hat und im Kriterium genannt wird, wohnt er nicht auerhalb. 
Ist der Wert im Kriterium nicht genannt, wohnt er auerhalb. In beiden Fllen werden die weiteren Kriterien 
nicht mehr betrachtet.
</li>
<li>Falls aufgrund fehlender Werte (oder fehlender Kriterien) kein Kriterium angewendet werden konnte, 
wird standardmig angenommen, dass der Klient nicht auerhalb wohnt. 
</li>
</ul>
Beispiele:
<ul>
<li>
<pre>
bezirk: Friedrichshain-Kreuzberg; ort: Berlin
</pre>
In Berlin enthalten alle mit dem Straenkatalog abgeglichene Adressen einen Wert fr den Bezirk. Da
jede Beratungsstelle genau fr einen Bezirk zustndig ist, kann auf dieser Basis entschieden werden.
Falls eine Adresse nicht mit dem Straenkatalog abgeglichen wurde, kann der Ort als Kriterium herangezogen 
werden, da in diesem Fall fr Bezirk keine Angaben vorliegen.
</li>
<li>
Falls kein Straenkatalog verwendet wird, kann nur der Ort, die Postleitzahl oder evt. der Planungsraum
(falls solche definiert sind und routinemig mit den Klientenadressen zusammen erfasst werden) als Kriterium
verwendet werden, z.B. (fiktiv!):
<pre>
plz: 12345;23456;34567
</pre>
Adressen mit diesen Postleitzahlen werden als nicht auerhalb wohnend betrachtet, alle
anderen als auerhalb. Falls die Klientenadresse keine 5-stellige PLZ enthlt, gilt der Wert als fehlend. 
Ungenauigkeiten werden in Kauf genommen, da die Region der Zustndigkeit nicht immer exakt mit
den Regionen der Postleitzahlen bereinstimmt.
</li>
<li>
<pre>
samtgemeinde: Name1;Name2;Name3; plz: 12345;23456
</pre>
Analog zum Berliner Beispiel wrde hier auch zunchst das Kriterium Samtgemeinde versucht, das
aber nur nach einem Abgleich mit dem Straenkatalog bekannt ist. Fr die nicht abgeglichenen Adressen
wird dann die Postleitzahl herangezogen. In seltenen Fllen kann das zu Fehlern fhren, da
die Bereiche der Postleitzahlen manchmal nicht mit den Bereichen der Zustndigkeit exakt bereinstimmen.
</li>
</ul>

Falls das Ergebnis "wohnt auerhalb" ist, wird automatisch ein entsprechender Eintrag in die
Bundesstatistik vorgenommen, der beim Ausfllen oben im Kasten "Falldaten" angezeigt wird.
Wenn mglich wird der <strong>amtliche Gemeindeschlssel</strong> eingetragen. Wenn dieser 
nicht gefunden wird, wird Postleitzahl und Wohnort angegeben.
<br><br>

<strong>Syntax</strong> wie in den Beispielen. Rund um die Trenner <code>;</code> und <code>:</code> drfen Leerzeichen stehen.
Es knnen auch mehrere Zeilen verwendet werden. 
<br><br>
Unter (Berater-) "Hauptmen -> Statistik -> Teilmenge -> Neu -> Neue Bedingung" finden Sie 
fr jedes Kriterium die Menge der mglichen Werte.
<br><br>

Zur <strong>Testen</strong> gibt es unter "Aministratorhauptmen -> Bundesstatistik" einen neuen 
Punkt "Adressen auerhalb prfen". Sie knnen dort in einer Tabelle sehen, welche Adressen als
auerhalb klassifiziert werden.
""",
         ),

    'gemeindeschluessel_von_plz':
        (False,
         r"^\s*$|^\s*\d{1,5}\s*(;\s*\d{1,5}\s*)*$",
         """Definiert die Teilmenge der amtlichen Gemeindeschlssel fr die Bundesstatistik.""",
         """In der Bundesstatistik soll der amtliche Gemeindeschlssel fr den Wohnort des Klienten 
eingetragen werden, wenn der Wohnort des Klienten auerhalb des Kreises der Beratungsstelle liegt. 
Um die Menge der Daten zu begrenzen, kann hier ber die Anfangsziffern der Postleitzahl eine Teilmenge 
definiert werden. Wenn mehrere Ziffernfolgen angegeben werden, mssen diese mit <code>;</code> getrennt 
werden.<br><br>
Im Normalfall brauchen Sie den voreingestellten Wert nicht zu ndern.
<br><br>
Beispiele:
<ul>
<li>Alle Gemeindeschlssel verwenden, auer die Postleitzahl der Gemeinde beginnt mit '9':
<pre>
0;1;2;3;4;5;6;7;8
</pre>
</li>
<li>Groe Teile Niedersachsens:
<pre>
29;30;31;37;38
</pre>
</li>
<li>Nrdliches Niedersachsen und Schleswig Holstein:
<pre>
2
</pre>
</li>
<li>Berlin und Umgebung:
<pre>
10;12;13;14;15;16
</pre>
</li>
<li>Bleibt die Variable leer, werden alle Gemeindeschlssel verwendet.
</li>
</ul>

Wenn es bei der Erkennung von Gemeindeschluesseln zu vielen Fehlern kommt, beispielsweise wegen
unterschiedlicher Schreibweisen von Orten, kann man selber eine Datei 
<code>EBKUS_HOME/sql/gemeindeschluessel.csv</code> erstellen, bestehend aus einer Tabelle mit den Spalten 
<code>plz, ort ags</code>. Als Vorlage kann die Datei
<code>EBKUS_HOME/sql/gemeindeschluessel_cache.csv</code> dienen, die automatisch erzeugt wird, 
wenn die Konfigurationsvariable <code>gemeindeschluessel_von_plz</code> einen Wert hat.
""",
         ),

    'neumeldungen_nach_region':
        (True,
         r"^\s*$|^\s*[^;]+\s*(;\s*[^;]+\s*)*$",
         """Definiert eine Tabelle "Neumeldungen nach Region" unter "Neumelde- und Abschlusszahlen".""",
         """Die Regionen, fr die in der Tabelle die Neumeldungen ausgezhlt werden sollen, mssen unter 
"Hauptmen -- Statistikabfrage -- Teilmenge -- Neu" als Teilmengendefinitionen eingefhrt werden. Der Namen der
gewnschten Definitionen werden durch <code>;</code> getrennt aufgefhrt.
<br><br>
Beispiel:
<ul>
<li>
<pre>
    Gifhorn und Samtgemeinde Meinersen; Braunschweig 38120 und 38112
</pre>
</li>
</ul>
"Gifhorn und Samtgemeinde Meinersen" ist eine Teilmengendefinition mit der Definition:
<code>( ort = 'Gifhorn' ODER samtgemeinde = 'Meinersen' )</code>.
"Braunschweig 38120 und 38112" ist eine Teilmengendefinition mit der Definition:
<code>( ort = 'Braunschweig' UND plz = '38112' oder '38120' )</code>.
<br><br>
""",
         ),
    
    'meldung_vom_admin':
        (True,
         "",
         """Inhalt wird als Meldung vom Administrator unbersehbar auf jeder EBKuS-Seite ber der Statuszeile ausgegeben.""",
         """Damit kann der Administrator
Meldungen an alle aktiven EBKuS-Nutzer ausgeben, z.B. um einen
Ausfall von EBKuS wegen Wartungsarbeiten anzukndigen.
""",
         ),
    
    'keine_bundesstatistik':
        (False,
         "",
         """Fall kann ohne Bundestatistik abgeschlossen werden.""",
         """Alles andere sollte trotzdem noch funktionieren.
""",
         ),

    'leistungsende_in_zukunft_erlaubt':
        (True,
         "",
         """Beim Eintrag einer Leistung darf fr das Ende ein Datum, das in der Zukunft liegt, eingetragen werden.""",
         """Der Abschluss des Falls ist erst mglich, wenn alle Leistungen beendet sind.
""",
         ),
    'cache_on':
        (True,
         "",
         """Falls true, wird der Objektcache eingeschaltet, sonst bleibt er abgeschaltet.""",
         """Normalerweise ist der Cache an, aber es kann in seltenen Fllen zu Fehlern kommen, die 
verschwinden, wenn er abgeschaltet wird. Auf die inhaltlichen Ablufe hat das keine Auswirkungen, 
allenfalls auf die Performance.
""",
         ),
    'jgh_gemeindeschluessel':
        (True,
         r"^$|^blank$|^\d{3,3}$",
         """Die letzten drei Ziffern des <a href="http://de.wikipedia.org/wiki/Amtlicher_Gemeindeschl%C3%BCssel" target="_blank">Amtlichen Gemeindeschlssels</a> oder das Wort "blank".""",
         """Standardmig setzt EBKuS fr diese drei Ziffern (Spalte 7-9 des Exportdatensatzes) drei mal die Ziffer 0 ein. Unseres Wissens werden diese Daten fr die
Bundesstatistik allerdings nicht ausgewertet. Bei Bedarf kann hier das Wort "blank" angegeben werden, dann erscheinen im Datensatz drei Leerzeichen, oder es knnen drei Ziffern
eingegeben werden, die dann in den Datensatz eingefgt werden.
""",
         ),
    'erinnern_nach_wochen_ohne_kontakt':
        (True,
         r"^[0123456789]\d{0,1}$",
         """Anzahl der Wochen ohne Kontakt nach der ein offener Fall in die Erinnerungsfunktion aufgenommen wird.""",
         """Wenn hier eine Zahl zwischen 1 und 99 steht, wird die Erinnerungsfunktion aktiviert. Auf einem zustzlichen Button im Hauptmen steht dann die Anzahl der offenen Flle, mit denen es lnger als die angegebene Anzahl von Wochen keinen Kontakt gegeben hat. Nach dem Bettigen des Buttons erscheint eine Liste dieser Flle. Jeder Fall kann als erledigt markiert werden, worauf er knftig normalerweiser nicht mehr in der Liste erscheint. Fr die gesamte Liste gibt es aber eine Option, alle Flle einschlielich der erledigten anzuzeigen. Die 'Erledigt'-Markierung kann dann auch wieder rckgngig gemacht werden.
<br><br>
Es muss hier die Ziffer 0 stehen, wenn die Erinnerungsfunktion deaktiviert bleiben soll.
""",
         ),
    'anmeldeprobleme_und_beschaeftigung_bei_anmeldung_erfassen':
        (True,
         "",
         """Falls true, werden Probleme bei der Anmeldung und Beschftigung schon bei der Anmeldung erfasst und spter fr die Fachstatistik bernommen.""",
         """Die Auswahlfelder 'Beschftigung Jugendliche(r)', 'Beschftigungsverhltnis Mutter', 'Beschftigungsverhltnis Vater' und das Mehrfachauswahlfeld 'Problem(e) bei der Anmeldung' aus der Fachstatistik werden in den Kasten Anmeldekontakt als Pflichtfelder aufgenommen. Es ist jedoch in jedem Fall mglich 'Keine Angabe' auszuwhlen (genau wie in der Fachstatistik selbst). Die Daten werden mit dem Fall (nicht mit der Akte bzw. den Stammdaten) gespeichert. Wenn eine Fachstatistik angelegt wird, werden die Daten bernommen.
""",
         ),
    'log_query_laenger_als':
        (True,
         r"^[0123456789]\d{0,5}$",
         """Falls grer 0, Datenbankqueries in die Logdatei schreiben, die lnger dauern als die angegeben Zahl in Millisekunden.""",
         """Wenn 0 eingetragen wird, werden keine Datenbankqueries protokolliert. Alle Datenbankqueries knnen mit log_level=DEBUG erfasst werden. Ntzlich zur Diagnose von Performanceproblemen.
""",
         ),
    'fallbeginn_waehlbar':
        (True,
         "",
         """Bei Neumelde- und Abschlusszahlen sowie Klienten pro Mitarbeiter kann ein Hkchen gesetzt werden, um den Fallbeginn ab erster Leistung zu definieren.""",
         "",
         ),
    'fallbeginn_bei_erster_leistung':
        (True,
         "",
         """Bei Neumelde- und Abschlusszahlen sowie Klienten pro Mitarbeiter wird der Fallbeginn ab dem Datum der erste Leistung definiert anstatt ab dem Anmeldedatum.""",
         "",
         ),
    'today':
        (False,
         "",
         """Datum, das EBKuS fr den heutigen Tag hlt, z.B. 2015-12-03.""",
         """Damit knnen reproduzierbare Tests durchgefhrt werden. Darf nur fr Demos bzw. Test verndert werden, in der Produktion natrlich nicht.
""",
         ),
    'anzahl_demoakten':
        (False,
         "",
         """Anzahl der Demoakten, die fr die Instanz bei Initialisierung der Datenbank angelegt werden.""",
         """Hat nur ein Effekt, wenn die Instanz eine Demoinstanz ist und bei der Erstanlage der Datenbank.
""",
         ),
    'sperrfrist':
        (True, 
         r"^\d{1,2}|-1$",
         "Falls der Wert positiv ist: Anzahl der Monate nach der die Akte eines abgeschlossenen Falls automatisch gesperrt wird. "
         "Falls der Wert 0 ist: keine automatische Sperrung, aber manuelle Sperrung mglich. "
         "Falls der Wert -1 ist: alle Sperrungen werden aufgehoben und der Sperren/Entsperren-Button im Administratormen wird nicht gezeigt.",

         "Unabhngig von der automatischen Sperrung kann der Anwendungsadministrator immer auch einzelne Akten manuell "
         "sperren bzw. entsperren, solange der Wert nicht auf -1 gesetzt wird. <br/>"
         "Die Klientenkarte einer gesperrten Akte kann nicht mehr geffnet "
         "werden. In der Auswertung \"Beratungen\" bleibt ein Eintrag "
         "erhalten, der Klientenname ist aber nicht lesbar. "
         "Bei Gruppen werden keine Teilnehmer aufgefhrt, deren Akte gesperrt ist. "
         "Die Daten einer gesperrten Akte gehen aber weiterhin in alle Statistiken ein. "
         "Bei der Suche knnen gesperrte Akten nicht gefunden werden. " 
         ),
    'beratungskontakt_dokumentation':
        (True,
         "",
         """Falls <code>true</code>, knnen Beratungskontakte vollstndig dokumentiert werden.""",
         """Vollstndige Dokumentation einer Beratung ist damit in EBKuS mglich. Dazu werden eine Reihe von 
zustzlichen Feldern angeboten, die fr die Dokumentation zur Verfgung stehen. 
""",
         ),
    'beratungskontakt_dokumentation_bearbeitungs_frist':
        (True,
         r"^\d{1,2}|-1$",
         """Anzahl der Werktage nach dem Ersteintrag, in denen noch eine Bearbeitung der Beratungsdokumentation erfolgen kann.""",
         """Danach ist sie nur noch lesbar. Falls "-1" eingetragen wird, bleibt sie permanent modifizierbar.
""",
         ),
    'beratungskontakt_dokumentation_persoenliches_erstgespraech_obligatorisch':
        (True,
         "",
         """Falls <code>true</code>, muss der der erste Beratungskontakt ein persnliches Gesprch sein.
""",
        """Hat nur dann einen Effekt, wenn die Variable <code>beratungskontakte_bs</code> auf <code>true</code> steht.
""",
         ),
    'wireb':
        (True,
         None,
         "Falls <code>true</code> wird die Teilnahme am Projekt Wir.EB untersttzt.",
         "Teilnahme am Projekt \"Wirkungsevaluation in der Erziehungsberatung\" (https://www.wireb.de/). "
         "Auf der Klientenkarte erscheint ein zustzlicher Kasten 'Wir.EB-Anbindung'. "
         "",
         ),
    'wireb_api_endpoint':
        (True,
         r'^(http|https):/.+[.][a-z]+/.*$',
         "Endpunkt der Wir.EB-API. Ermglich aus EBKuS heraus Anfragen an Wir.EB.",
         "Achtung: Diese Adresse muss vom EBKuS-Server aus netztechnisch erreichbar sein, sonst kann die Wir.EB-Einbindung nicht funktionieren.",
         ),
    'wireb_statistik_in_endpoint':
        (True,
         r'^(http|https):/.+[.][a-z]+/.*$',
         "Endpunkt fr das Hochladen der Bundesstatistik an Wir.EB.",
         "Achtung: Diese Adresse muss vom EBKuS-Server aus netztechnisch erreichbar sein, sonst kann die Bundesstatistik nicht hochgeladen werden.",
         ),
    'wireb_beratungsstelle':
        (True,
         '',
         "Beratungstellennummer (Ganzzahl). Wird von Wir.EB vergeben und ist hier einzutragen.",
         "",
         ),
    'wireb_anbieter':
        (True,
         "",
         "Kennung fr Softwareanbieter (Ganzzahl). Fr EBKuS ist dies '5'.",
         "",
         ),
    'wireb_web_user':
        (True,
         "",
         "Von Wir.EB zugewiesener Benutzername fr die manuelle Anmeldung bei Wir.EB-Online (http://wireb.ikj-online.de/).",
         "Diesen Benutzernamen muss der EBKuS-Anwender selbst verwenden, um einen Fall bei Wir.EB anzumelden und um bei Wir.EB die Fallbgen auszufllen."
         "Erscheint im Tooltip fr dem Link zur Erinnerung.",
         ),
    'wireb_web_password':
        (True,
         "",
         "Von Wir.EB zugewiesenes Passwort fr die manuelle Anmeldung bei Wir.EB-Online (http://wireb.ikj-online.de/)",
         "Dieses Passwort muss der EBKuS-Anwender in Verbindung mit dem Benutzernamen selbst verwenden, "
         "um einen Fall bei Wir.EB anzumelden und um bei Wir.EB die Fallbgen auszufllen. "
         "Erscheint im Tooltip fr dem Link zur Erinnerung."
         ),
    'wireb_api_user':
        (True,
         "",
         "Von Wir.EB zugewiesener Benutzername fr die Wir.EB-API.",
         "Wird von der EBKuS-Software zur direkten Kommunikation mit der Wir.EB-API bentigt. Der EBKuS-Anwender benutzt diese Kennung niemals direkt selbst.",
         ),
    'wireb_api_password':
        (True,
         "",
         "Von Wir.EB zugewiesenes Passwort fr die Wir.EB-API.",
         "Wird von der EBKuS-Software zur direkten Kommunikation mit der Wir.EB-API bentigt. Der EBKuS-Anwender benutzt dieses Passwort niemals direkt selbst.",
         ),
    'zusaetzliche_berechtigungen':
        (True,
         None,
         "Falls <code>true</code> "
         "kann auf der Klientenkarte weiteren Mitarbeitern auer dem aktuell Zustndigen "
         "der Zugriff auf die Klientenkarte erlaubt werden.",
         "Bei den zustzlich Berechtigten erscheint dann der betreffende Klient "
         "bei der Klientenauswahl zustzlich zu denen, fr die er selber zustndig ist. "
         "Auerdem auch in der Liste der abgeschlossenen und laufenden Beratungen unter "
         "Auswertungen im Hauptmen.",
         ),
    'berichtseinheitid':
        (True,
         r"^(\s*Stelle\s+([^:]+)\s*:\s*(\d{5}( {3}|\d{3})\d{6})\s*;\s*)+$|^(\d{5}( {3}|\d{3})\d{6})$",
         "Die BerichtseinheitID fr die Einrichtung, die zur Abgabe der Jugendhilfestatistik verpflichtet ist.",
         "Die BerichtseinheitID besteht aus genau 14 Ziffern, die wie folgt zusammengesetzt sind: "
         "Stelle 1-8 ist der Gemeindeschlssel (Land 1-2, Regierungsbezirk 3, Kreis 4-5, Gemeinde 6-8) und "
         "die Stellen 9-14 ist die Einrichtungsnummer. <br>"
         "Die drei Stellen fr die Gemeinde wurden bisher manchmal leer gelassen, evt. ist das auch weiter mglich. "
         "In diesem Fall an Stelle der 3 Ziffern fr die Gemeinde genau drei Leerzeichen eingeben."
         "<br><br>"
         "Falls die BereichtseinheitID fr alle von der EBKuS-Instanz abgedeckten Stellen dieselbe ist, gengt es, allein "
         "diese einzugeben: "
         "<pre>"
         "12345678123456"
         "</pre>"
         "Falls die Stellen verschiedene BerichtseinheitIDs haben, bitte das folgende Format verwenden: "
         "<pre>"
         "Stelle A: 12345678123456; Stelle B: 12345678234567; Stelle C: 12345678345678;"
         "</pre>"
         "das heit fr jede Stelle die Folge das Wort 'Stelle', Leerzeichen, Stellenkrzel, Doppelpunkt, 14-stellige BerichtseinheitID, Semikolon eingeben.",
         ),
    
    }

if __name__ == "__main__":
    import doctest
    doctest.testmod()

