# coding: latin-1
import os, sys, logging
import re
import datetime
from os.path import join, exists
from time import time
from itertools import chain
from ebkus import Version
from ebkus.config import config
from ebkus.db.sql import opendb, closedb, getDBHandle, SQL, SQLError
from ebkus.db.dbapp import DBAppError
from ebkus.app.ebapi import EE, Tabelle, Feld, FeldList, \
     Kategorie, KategorieList, Code, CodeList, FachstatistikList, \
     FallList, AkteList, JugendhilfestatistikList, Jugendhilfestatistik2007List, \
     getNewFallnummer, cc, today, calc_age, Fall, \
     register_get, register_set, get_berichtseinheitid
from ebkus.app.ebapih import get_codes, get_all_codes
from ebkus.app.jghdef import get_jghdef, get_bundesstatistikmanager
from ebkus.utils import dtstr_from_dt, getmtime_as_dtstr, dt_from_dtstr


"""
Alle Updates sollten so geschrieben werden, dass sie problemlos
wiederholt werden knnen, ohne dass etwas schief geht.
Zumindest sollte das fuer alle Updates ab heute gelten :-)

Wenn die Version als letzte Komponente '.devel' enthlt, wird das erfolgreiche 
Update nicht in das Register eingetragen, so dass das Update immer wieder
wiederholt wird.

"""

def update():
    u = UpdateDB()
    # hack: knstlich auf 4.2 setzen, nur fuer demo_test
    #from ebkus.app.ebapi import register_set
    #register_set("Version", "4.2")
    #u.ist = u.get_version()
    u.keep_alive_anpassen()
    u.version_module_laden()
    # Vorlufig soll das immer laufen, auch wenn schon 4.12 als DB-Version eingefhrt ist
    # Fuer die Entwicklung.
    # Spter wieder rausnehmen
    u.add_zusaetzliche_berechtigung()
    # DARF hier nicht passieren, wenn das Schema gendert wurde, da Objekte
    # mit dem neuen Code nicht richtig aus der alten DB gelesen werden knnen!
    #u.fsqualij_code_reparieren()
    if u.update_noetig():
        u.update()
    # wireb_bei_bedarf_nachtragen() kann spter wieder raus, evt. noch bei 4.9 auf 4.10 drin lassen
    # Sollte immer gelingen, da geprft wird, ob die Vernderungen schon da sind.
    u.wireb_bei_bedarf_nachtragen()
    u.indexe_bei_bedarf_einfuehren()
    u.fallnummer_pruefen_und_reparieren()
    # Update der Bundesstatistik nach jedem Neustart, da diese sich ndern kann auch ohne
    # eine neue EBKuS-Version
    u.update_bundesstatistikdefinition()
    
class UpdateDB(object):
    """Zum Update einer EBKuS-Datenbank von einer Version auf eine hhere.
    Die Software der hheren Version muss bereits eingespielt sein.
    """

    # Die Software kann auch immer eine dreistellige Version sein (4.3.1).
    # Die zugehrige DB-Version ist immer zweistellig. 
    # Wenn das Schema gendert wird, muss sich die zweistellige Versionszahl
    # erhhen.

    # Die Datenbank muss mindestens auf diesem Niveau sein, damit 
    # ein Update mglich ist.
    ist_moeglich_db = ['4.10', '4.11', '4.12']
    # Die Update Software kann die Datenbankversion auf dieses Niveau
    # heben (d.h. es gibt eine update_<ist_mglich>_nach_<soll_mglich>
    soll_moeglich_db = ['4.13']

    varchar_regex = re.compile(r'VARCHAR\((\d+)\)')

    def __init__(self):
        # Protokoll stoert
        from ebkus.app.protocol import temp_off
        temp_off()
        # Sollversion ist immer die Major-Version der Software (4.1, 4.4, ...)
        devel_indicator = '.devel'
        if Version.endswith(devel_indicator):
            self.devel = True 
        else:
            self.devel = False
        # Minor-Version spielt keine Rolle
        # self.soll_db ist die Major-Version der aktuellen Version
        vt = Version.split('.')
        self.soll_db = "%s.%s" % (vt[0], vt[1])
        try:
            opendb()
        except:
            # Zum Update muss die Datenbank verfgbar sein.
            # Daher:
            self.abort("Fuer das Update muss die Datenbank verfuegbar sein!",
                       exc_info=True)
        # Ist-Version der Datenbank feststellen
        self.tables = getDBHandle().listtables()
        try:
            ist_db = self.get_version()
            vt = ist_db.split('.')
            self.ist_db = "%s.%s" % (vt[0], vt[1])
        except:
            self.ist_db = None

    def abort(self, message, exc_info=False):
        logging.critical(message, exc_info=exc_info)
        logging.shutdown()
        sys.exit(1)

    def has_field(self, table, field):
        return bool([t for t in getDBHandle().listfields(table) if t[0] == field])

    def get_version(self):
        tables = self.tables
        if ('jghstat07' in tables and
            not 'strkatalog' in tables and
            'fachstatlei' in tables):
            return '3.3'
        elif 'register' in tables:
            try:
                version = register_get('Version')
                if version:
                    return version
            except SQLError:
                pass
            try:
                # Neues Schema ab 4.4, DB ist lter, also Versionsinfo direkt mit SQL holen
                werte = SQL("SELECT value FROM register where regkey='Version'").execute()
                if werte:
                    from cPickle import loads
                    s = werte[0][0]
                    try: s = s.tostring() # kommt manchmal als array.array raus
                    except: pass # war schon string
                    res = loads(s)
                    return res
            except:
                pass
            # register-Tabelle vorhanden aber keine Version gefunden, also 4.0 oder 4.0.1
            return '4.0'
        else:
            return None

    def update_noetig(self):
        if not self.ist_db:
            logging.error("Version der Datenbank nicht feststellbar, Update nicht mglich.")
            return False
        elif self.ist_db == self.soll_db:
            logging.info("Datenbankversion und Softwareversion sind kompatibel.")
            logging.info("Kein Update erforderlich")
            return False
        else:
            logging.info("Datenbankversion: %s" % self.ist_db)
            logging.info("Softwareversion:  %s" % Version)
            logging.info("Update der Datenbank erforderlich")
        if self.soll_db not in self.soll_moeglich_db:
            self.abort("Update auf Version %s nicht moeglich" % Version)
        if not self.ist_db in self.ist_moeglich_db:
            self.abort("Update von Version %s nicht moeglich" % self.ist_db)
        logging.info("Datenbank wird auf Version %s updatet" % self.soll_db)
        return True

    def update(self):
        method = "update_%s_nach_%s" % (self.ist_db, self.soll_db)
        method = method.replace('.', '_')
        res = getattr(self, method)()
        from ebkus import Version
        if self.devel:
            logging.info("Datenbank von Version {} auf {} updatet"
                         ", aber kein Registereintrag, damit das Update wiederholt werden kann.".format(
                             self.ist_db, self.soll_db))
        else:
            register_set("Version", self.soll_db)
        logging.info("Datenbank von Version %s auf Version %s erfolgreich updatet" %
                     (self.ist_db, self.soll_db))

    # 4.11
    def update_4_10_nach_4_11(self):
        self.add_field('beratungskontakt', 'pervid', 'per Video', 'INT', 'k', kat_code='ja_nein')
    # 4.12
    def update_4_11_nach_4_12(self):
        self.add_kategorie('inobh42', 'Grund der Inobhutnahme', '', 0)
        self.add_code('inobh42', '1',
                      'Ja, aufgrund einer unbegleiteten Einreise aus dem Ausland '
                      '( 42 Absatz 1 Satz 1 Nummer 3 SGB VIII)',
                      1)
        self.add_code('inobh42', '2',
                      'Ja, aufgrund einer dringenden Kindeswohlgefhrdung '
                      '( 42 Absatz 1 Satz 1 Nummer 2 SGB VIII)',
                      2)
        self.add_code('inobh42', '3',
                      'Ja, aufgrund der Bitte des Kindes/Jugendlichen um Inobhutnahme '
                      '( 42 Absatz 1 Satz 1 Nummer 1 SGB VIII)',
                      3)
        self.add_code('inobh42', '4',
                      'Nein, zuvor wurde keine Inobhutnahme durchgefhrt',
                      4)
        self.add_field('jghstat07', 'inobh42_22', 'Nach Inobhutnahme gem 42 Absatz 1 Satz 1 SGB VIII', 'INT', 'k', kat_code='inobh42')
        # neues Feld initialisieren aus den Wert von inobh42
        # wenn inobh42 auf 1 steht (unbegleitete Einreise ja) --> inobh42_22 auf 1, falls vorher null
        # wenn inobh42 auf 2 steht (unbegleitete Einreise nein) --> inobh42_22 auf 4, falls vorher null
        SQL("""UPDATE jghstat07 SET inobh42_22=%s
WHERE inobh42_22 IS NULL AND inobh42 =%s""" % (cc('inobh42', '1'), cc('ja_nein', '1'))).execute()
        SQL("""UPDATE jghstat07 SET inobh42_22=%s
WHERE inobh42_22 IS NULL AND inobh42 =%s""" % (cc('inobh42', '4'), cc('ja_nein', '2'))).execute()
        logging.info("Werte aus dem alten Item Inobhutnahme in der Bundesstatistik "
                     "in das neue veraenderte Item uebernommen, wie folgt:\n"
                     "  1. Uebernahme nur dann, wenn altes Item einen Wert hat und das neue Item keinen.\n"
                     "  2. Wenn im alten Item 'unbegleitete Einreise' bejaht wurde, Uebernahme entsprechend nach Alternative 1.\n"
                     "  3. Wenn im alten Item verneint wurde, Uebernahme nach Alternative 4, 'keine Inobhutnahme'.")
        # Orte daktivieren, die fuer EB 28 nicht in Frage kommen
        self.activate_code('hilf_ort', '03', 1)
        self.activate_code('hilf_ort', '07', 1)
        self.activate_code('hilf_ort', '08', 1)
        self.activate_code('hilf_ort', '09', 1)
        self.activate_code('hilf_ort', '10', 1)
        self.activate_code('hilf_ort', '11', 1)
        self.add_code('hilf_ort', '13', 'Per Telefon', sort=10)
        self.add_code('hilf_ort', '14', 'ber das Internet (z.B. Chatberatung, Videokonferenz)', sort=11)
        self.rename_code('shf', '2', 'Elternteil lebt alleine ohne (Ehe-)Partner/in')
        self.rename_code('shf', '3', 'Elternteil lebt mit neuer Partnerin/neuem Partner (z.B. Stiefelternkonstellation)')
        self.rename_code('shf', '4', 'Eltern sind verstorben')
        self.rename_code('shf', '5', 'Unbekannt')
        self.rename_code('unh', '2', 'Weiterverweisung an Eheberatung, Schuldnerberatung, Kinder- und Jugendlichenpsychotherapeuten, andere Einrichtungen ( 17-21 SGB VIII)')
        self.rename_code('unh', '6', 'Keine unmittelbar nachfolgende Hilfe nach dem SGB VIII bekannt')
        self.rename_code('auf_ort', '03', 'In einer nicht.verwandten Familie (z.B. Pflegestelle nach  44 SGB VIII)')
        self.rename_code('auf_ort', '05', 'In einer Pflegefamilie nach  33, 35a, 41 SGB VIII')
        self.rename_code('auf_ort', '06', 'In einem Heim oder einer betreuten Wohnform nach  34, 35a, 41 SGB VIII')
        self.rename_code('auf_ort', '10', 'Ohne feste Unterkunft')
        self.rename_code('auf_ort', '11', 'Unbekannt/keine Angabe mglich')
        self.add_zusaetzliche_berechtigung()
    def add_zusaetzliche_berechtigung(self):
            self.add_table('zusaetzliche_berechtigung',
                           'Zusaetzlich Berechtigte zur Bearbeitung des Falles',
                           'ZusaetzlicheBerechtigung')
            self.add_field('zusaetzliche_berechtigung', 'fall_id', 'Fallid', 'INT', 'f',
                           'zusaetzlich_berechtigte', None, 'fall')
            self.add_field('zusaetzliche_berechtigung', 'mit_id', 'Mitarbeiterid', 'INT', 'f',
                           'zusaetzlich_berechtigt_fuer', None, 'mitarbeiter')
            self.add_field('beratungskontakt', 'dvon_id', 'Mitarbeiterid', 'INT', 'f',
                           'erstellte_dokus', None, 'mitarbeiter')
            self.add_field('beratungskontakt', 'dlcvon_id', 'Mitarbeiterid', 'INT', 'f',
                           'geaenderte_dokus', None, 'mitarbeiter')

    def update_4_10_nach_4_12(self):
        self.update_4_10_nach_4_11()
        self.update_4_11_nach_4_12()
    # 4.13
    def update_4_12_nach_4_13(self):
        self.text_felder_verlaengern()
        self.set_berichtseinheitid()
        self.add_field('akte', 'abwlf', 'Abweichende Lschfrist in Monaten', 'INT', 'p',
                       dok='Abweichende Lschfrist in Monaten seit dem Ende des letzten Falls')
        self.add_joker_fuer_deklarative_bundesstatistik()
    def set_berichtseinheitid(self):
        try:
            # config hier noch nicht richtig initialisiert, daher register_get statt config.BERICHTSEINHEITID
            stz_berichtseinheitid = get_berichtseinheitid(config_string=register_get('berichtseinheitid'))
            for stz_code, berichtseinheitid in stz_berichtseinheitid.items():
                logging.info("BerichtseinheitID '{}' fuer Stelle '{}' gefunden.".format(berichtseinheitid, stz_code))
            return
        except Exception:
            # nicht oder falsch konfiguriert
            # raise
            pass
        # nach Stellenzeichen getrennt info aus land, kr, einrnr Kategorie holen, falls vorhanden
        # get legacy
        from ebkus.html.options import get_land_kr_einrnr
        def get_part_code(kat_code, stelle):
            # holt den code fuer eine Kategorie und eine Stelle
            # print "CODE FUER:", kat_code, stelle['code']
            code = get_land_kr_einrnr(kat_code, stelle)
            # print "GEFUNDEN:", code
            if code:
                return code['code']
            else:
                return ''
        stz_berichtseinheitid = {}
        for stelle in get_codes('stzei'):
            # land, kr, gem, einrnr
            land = get_part_code('land', stelle)
            kr = get_part_code('kr', stelle)
            einrnr = get_part_code('einrnr', stelle)
            gem = '   '
            jgh_gemeindeschluessel = config.JGH_GEMEINDESCHLUESSEL
            if jgh_gemeindeschluessel.isdigit() and len(jgh_gemeindeschluessel) == 3:
                gem = jgh_gemeindeschluessel
            berichtseinheitid = land + kr + gem + einrnr
            stz_berichtseinheitid[stelle['code']] = berichtseinheitid
        def eine_berichtseinheitid_fuer_alle():
            x = None
            nur_eine = True
            for berichtseinheitid in stz_berichtseinheitid.values():
                if x is None:
                    x = berichtseinheitid
                else:
                    if x != berichtseinheitid:
                        nur_eine = False
            return nur_eine
        if eine_berichtseinheitid_fuer_alle():
            config_string = stz_berichtseinheitid.values()[0]
        else:
            config_string = ''
            for stz_code, berichtseinheitid in stz_berichtseinheitid.items():
                config_string += "Stelle {}: {}; ".format(stz_code, berichtseinheitid)
        # Testen, ob die zusammengeklaubten berichtseinheitids korrekt sind
        try:
            stz_berichtseinheitid_test = get_berichtseinheitid(config_string=config_string)
            register_set('berichtseinheitid', config_string)
            logging.info(
                "Konfigurationsvariable 'berichtseinheitid' gesetzt mit dem Wert '{}'.".format(config_string))
        except EE as ee:
            logging.error("Keine korrekte Konfiguration fr die BerichtseinheitID gefunden: {}. "
                          "Bitte unter 'Admin -> Einstellungen -> berichtseinheitid' nachtragen!".format(ee.message))
        # # Keine gefunden
        # values = list(stz_berichtseinheitid.values()) # python 2 ist list nicht ntig
        # if not values:
        #     logging.error("Keine BerichtseinheitID gefunden. Bitte unter "
        #                   "'Admin -> Einstellungen -> berichtseinheitid' nachtragen!")
        #     return
        # if len(values) == 1:
        #     # es gibt nur eine Stellen bzw. alle Stellen haben die gleiche berichtseinheitid
        #     register_set('berichtseinheitid', values[0])
        #     logging.info(
        #         "BerichtseinheitID '{}' gefunden und unter 'Admin -> "
        #         "Einstellungen -> berichtseinheitid' eingetragen.".format(values[0]))
        # else:
        #     register_value = ''
        #     for stz_code, berichtseinheitid in stz_berichtseinheitid.items():
        #         register_value += "Stelle {}: {}; ".format(stz_code, berichtseinheitid)
        #     register_set('berichtseinheitid', register_value)
        #     for stz_code, berichtseinheitid in stz_berichtseinheitid.items():
        #         logging.info("BerichtseinheitID '{}' fuer Stelle '{}' gefunden und unter "
        #                       "'Admin -> Einstellungen -> berichtseinheitid' eingetragen.".format(
        #                           berichtseinheitid, stz_code))
    def add_joker_fuer_deklarative_bundesstatistik(self):
        # Kategorien fr Joker 4 Stck
        for i in range(1, 5):
            self.add_kategorie('jokkc{}'.format(i), 'Dummy')
        for i in range(1, 5):
            self.add_code('jokkc{}'.format(i), '1', 'Dummy')
        # Bereichskategorien fr Joker 2 Stck
        for i in range(1, 3):
            self.add_kategorie('jokbk{}'.format(i), 'Dummy')
        for i in range(1, 3):
            self.add_code('jokbk{}'.format(i), '1', 'Dummy', mini=1, maxi=999)
        # Felder fr Joker typ k
        for i in range(1, 5):
            self.add_field('jghstat07',
                           'jokk{}'.format(i),
                           'Joker k {}'.format(i),
                           'INT', 'k',
                           kat_code='jokkc{}'.format(i))
        # Felder fr Joker typ b int
        for i in range(1, 3):
            self.add_field('jghstat07',
                           'joki{}'.format(i),
                           'Joker int {}'.format(i),
                           'INT', 'b',
                           kat_code='jokbk{}'.format(i))
        # Felder fr Joker typ p text
        for i in range(1, 3):
            self.add_field('jghstat07',
                           'jokt{}'.format(i),
                           'Joker text {}'.format(i),
                           'VARCHAR(255)', 'p',)
            
    def update_4_11_nach_4_13(self):
        self.update_4_11_nach_4_12()
        self.update_4_12_nach_4_13()
    def update_4_10_nach_4_13(self):
        self.update_4_10_nach_4_12()
        self.update_4_12_nach_4_13()
    
    def add_kategorie(self, kat_code, name, dok='', flag=0):
        try:
            kat = Kategorie(code=kat_code)
            logging.info("Kategorie '%s' existiert bereits" % kat_code)
            return
        except DBAppError:
            pass # existiert nicht, weitermachen
        # insert
        c = Kategorie()
        c.new()
        c.init(
            code=kat_code,
            name=name,
            dok=dok,
            flag=flag,
            zeit=int(time()),
            )
        c.insert()
        logging.info("Kategorie '%s' hinzugefuegt" % (kat_code,))
        
        
    def add_code(self, kat_code, code, name, sort=1, dok='', mini=None, maxi=None):
        # auf Feldlnge achten, name ist bei alten Versionen nur 60 lang
        # test if kat_code exists
        try:
            kat = Kategorie(code=kat_code)
            kat_id = kat['id']
        except DBAppError:
            logging.error("Kategorie '%s' existiert nicht" % kat_code)
            raise
        # test if code already exists
        try:
            Code(kat_code=kat_code, code=code)
            logging.info("Code '%s' fuer Kategorie '%s' existiert bereits" % (code, kat_code))
            return
        except DBAppError:
            pass
        # insert
        c = Code()
        c.new()
        c.init(
            code=code,
            name=name,
            kat_id=kat_id,
            kat_code=kat_code,
            sort=sort,
            off=0,
            zeit=int(time()),
            mini=mini,
            maxi=maxi,
            )
        c.insert()
        logging.info("Code '%s' fuer Kategorie '%s' hinzugefuegt: %s" % (code, kat_code, name))

    def rename_code(self, kat_code, code, new_name, sort=None, dok=None):
        """sort und dok werden genau dann updatet, wenn sie nicht None sind."""
        # test if kat_code exists
        try:
            kat = Kategorie(code=kat_code)
            kat_id = kat['id']
        except DBAppError:
            logging.error("Kategorie '%s' existiert nicht" % kat_code)
            return
        # test if code exists
        try:
            c = Code(kat_code=kat_code, code=code)
        except DBAppError:
            logging.error("Code '%s' fuer Kategorie '%s' existiert nicht" % (code, kat_code))
            return
        # rename
        old_name = c['name']
        old_sort = c['sort']
        old_dok = c['dok']
        update_dict = dict()
        if new_name != old_name:
            update_dict = {'name': new_name}
        if sort is not None and sort != old_sort:
            update_dict['sort'] = int(sort)
        if dok is not None and dok != old_dok:
            update_dict['dok'] = dok
        c.update(update_dict)
        if 'name' in update_dict:
            logging.info("Code '%s' fuer Kategorie '%s' umbenannt: %s --> %s " % (code, kat_code, old_name, new_name))
        if 'sort' in update_dict:
            logging.info("Code '%s' fuer Kategorie '%s' Sortierung: %s --> %s " % (code, kat_code, old_sort, sort))
        if 'dok' in update_dict:
            logging.info("Code '%s' fuer Kategorie '%s' Doku: %s --> %s " % (code, kat_code, old_dok, dok))

    def activate_code(self, kat_code, code, off, dy=None, dm=None):
        try:
            c = Code(kat_code=kat_code, code=code)
        except DBAppError:
            logging.error("Code '%s' fuer Kategorie '%s' existiert nicht" % (code, kat_code))
            return
        assert off in (0, 1)
        if off == c['off']:
            return
        c.update({'off': off,
                  'dy': dy,
                  'dm': dm})
        prefix = 'de' if off == 1 else ''
        logging.info("Code '%s' fuer Kategorie '%s' %saktiviert" % (code, kat_code, prefix))
        
    def add_date(self, tabelle, name, datumart):
        """Datum bestehend aus Tag (d), Monat (m) und Jahr (y) im bewhrten Stil hinzufgen"""
        self.add_field(tabelle, name + 'd', 'Tag von ' + datumart, 'INT', 'p')
        self.add_field(tabelle, name + 'm', 'Monat von ' + datumart, 'INT', 'p')
        self.add_field(tabelle, name + 'y', 'Jahr von ' + datumart, 'INT', 'p')
    def add_field(self, tabellenname, name, langname, 
                  typ, verwtyp, 
                  inversename=None, kat_code=None, fremdtabellenname=None, flag=0, dok=''):
        """Feld zu einer bestehenden Tabelle hinzufgen.
        Wirft eine Exception
        - wenn ein Feld mit demselben Namen schon in der Tabelle ist
        - die Feldtabelle schon einen Eintrag mit diesem Namen hat
        - ein verwendeter verwtyp, kat_code oder Fremdtabelle nicht korrekt ist
        """
        tab_id = Tabelle(tabelle=tabellenname)['id']
        if self.has_field(tabellenname, name):
            fobj =  Feld(feld=name, tab_id=tab_id)
            # ok. Feld existiert in der Tabelle und hat einen Eintrag in Feld()
            logging.info("Feld '%s' fuer Tabelle '%s' existiert bereits" % (name,tabellenname))
            return
        kat_id = None
        if kat_code:
            kat_id = Kategorie(code=kat_code)['id']
        ftab_id = None
        if fremdtabellenname:
            ftab_id = Tabelle(tabelle=fremdtabellenname)['id']
        verwtyp_id = cc('verwtyp', verwtyp)
        if verwtyp in ('k', 'b', 'm'):
            assert kat_code, "Kategorie fehlt fuer Feld '%s'" % name
        if kat_code:
            assert verwtyp in ('k', 'b', 'm'), "Falscher verwtyp '%s' fuer Feld '%s'" % (verwtyp, name)
        SQL("ALTER TABLE %s ADD %s %s" % (tabellenname, name, typ)).execute()
        f = Feld()
        f.init(
            tab_id=tab_id,
            feld=name,
            name=langname,
            inverse=inversename,
            typ=typ,
            verwtyp=verwtyp_id,
            kat_id=kat_id,
            kat_code=kat_code,
            ftab_id=ftab_id,
            flag=flag,
            dok=dok,
            )
        f.new()
        f.insert()
        logging.info("Feld '%s' fuer Tabelle '%s' eingefuehrt" % (name, tabellenname))
        

    def text_felder_verlaengern(self):
        """
table, akte, , Akte, Akte
no, VARCHAR(255), Notiz, p,
        """
        self.varchar_feld_verlaengern('akte', 'no', 4096)
        self.varchar_feld_verlaengern('anmeldung', 'no', 4096)
        self.varchar_feld_verlaengern('einrichtung', 'no', 4096)
        self.varchar_feld_verlaengern('bezugsperson', 'no', 4096)
        self.varchar_feld_verlaengern('beratungskontakt', 'no', 4096)
        self.varchar_feld_verlaengern('fachstat', 'no', 4096)
        self.varchar_feld_verlaengern('fachstat', 'no2', 4096)
        self.varchar_feld_verlaengern('fachstat', 'no3', 4096)
        self.varchar_feld_verlaengern('fua_bs', 'no', 4096)
        self.varchar_feld_verlaengern('code', 'name', 4096)
        self.varchar_feld_verlaengern('code', 'dok', 4096)
        self.varchar_feld_verlaengern('kategorie', 'name', 4096)
        self.varchar_feld_verlaengern('kategorie', 'dok', 4096)
        self.varchar_feld_verlaengern('tabelle', 'name', 4096)
        self.varchar_feld_verlaengern('tabelle', 'dok', 4096)
        self.varchar_feld_verlaengern('feld', 'name', 4096)
        self.varchar_feld_verlaengern('feld', 'dok', 4096)
        
    def varchar_feld_verlaengern(self, tabellenname, name, length):
        """
        VARCHAR verlaengern.
        """
        assert length < 65536, "Feldlaenge zu gro: {}".format(length)
        tab_id = Tabelle(tabelle=tabellenname)['id']
        try:
            fobj =  Feld(feld=name, tab_id=tab_id)
                   # ok. Feld existiert in der Tabelle und hat einen Eintrag in Feld()
        except DBAppError:
            logging.error("Feld '%s' fuer Tabelle '%s' existiert nicht" % (name,tabellenname))
            return
        # rename
        old_typ = fobj['typ']
        match_old_typ = self.varchar_regex.match(old_typ)
        if not match_old_typ:
            logging.error("Feld '%s' fuer Tabelle '%s': Nur VARCHAR-Felder "
                          "duerfen veraendert werden" % (name,tabellenname))
        old_length = int(match_old_typ.groups()[0])
        if length == old_length:
            return
        if length < old_length:
            logging.error("Feld '%s' fuer Tabelle '%s': VARCHAR-Felder "
                          "duerfen nur verlaengert werden" % (name,tabellenname))
            return
        new_typ = "VARCHAR({})".format(length)
        update_dict = dict(typ=new_typ)
        SQL("ALTER TABLE %s MODIFY %s %s" % (tabellenname, name, new_typ)).execute()
        fobj.update(update_dict)
        logging.info("Das Feld '%s' in Tabelle '%s' auf %s verlaengert" %
                     (name, tabellenname, length))
    def rename_field_langname(self, tabellenname, name, langname):
        """
        Den Langnamen des Feldes in der Tabelle feld veraendern.
        """
        try:
            assert len(langname) < 4096
        except AssertionError:
            logging.error("Langname fuer Feld '%s' fuer Tabelle '%s' zu lang: %s" % (name,tabellenname, langname))
            raise
        tab_id = Tabelle(tabelle=tabellenname)['id']
        try:
            fobj =  Feld(feld=name, tab_id=tab_id)
        except DBAppError:
            logging.error("Feld '%s' fuer Tabelle '%s' existiert nicht" % (name,tabellenname))
            return
        old_langname = fobj['name']
        if old_langname != langname:
            update_dict = dict(name=langname)
            fobj.update(update_dict)
            logging.info("(Lang)name von Feld '%s' in Tabelle '%s' veraendert: %s" %
                         (name, tabellenname, langname))

    def keep_alive_anpassen(self):
        try:
            changed = False
            if sys.platform == 'win32':
                httpd_conf = join(config.INSTALL_DIR, 'apache', 'conf', 'httpd.conf')
                if exists(httpd_conf):
                    f = open(httpd_conf, 'rb')
                    text = f.read()
                    f.close()
                    if text.find('\nKeepAlive On') > -1:
                        text = text.replace('\nKeepAlive On', '\nKeepAlive Off')
                        open(httpd_conf, 'wb').write(text)
                        changed = True
            if changed:
                logging.info("Apache-Server auf 'KeepAlive Off' eingestellt")
        except:
            pass # sollte nie scheitern
    def version_module_laden(self):
        """Sort dafr, dass version_module unter Windows geladen wird."""
        if not sys.platform == 'win32':
            return
        try:
            found = changed = False
            httpd_conf = join(config.INSTALL_DIR, 'apache', 'conf', 'httpd.conf')
            httpd_conf22 = join(config.INSTALL_DIR, 'apache2.2', 'conf', 'httpd.conf')
            httpd_conf24 = join(config.INSTALL_DIR, 'apache2.4', 'conf', 'httpd.conf')
            httpd_conf24A = join(config.INSTALL_DIR, 'Apache24', 'conf', 'httpd.conf')
            for filename in (httpd_conf, httpd_conf22, httpd_conf24, httpd_conf24A):
                if exists(filename):
                    found = True
                    f = open(filename, 'rb')
                    text = f.read()
                    f.close()
                    if text.find('\n#LoadModule version_module') > -1:
                        text = text.replace('\n#LoadModule version_module', '\nLoadModule version_module')
                        open(filename, 'wb').write(text)
                        changed = True
            if found:
                if changed:
                    logging.info("Apache-Server: 'LoadModule version_module...' eingeschaltet durch Entfernen des Kommentarzeichens")
            else:
                logging.error("Apache-Server: 'httpd.conf' nicht gefunden")
        except:
            pass # sollte nie scheitern
            
    def fallnummer_pruefen_und_reparieren(self):
        faelle = FallList(where='')
        fn_dict = {}
        doubletten = []
        for f in faelle:
            fn = f['fn']
            count = fn_dict.setdefault(fn, 0)
            if count > 0:
                doubletten.append(fn)
            fn_dict[fn] += 1
        if doubletten:
            for fn in doubletten:
                n = fn_dict[fn]
                logging.error("****Fallnummer mehrfach vergeben: %s (%sx)" % (fn, n))
                self.fallnummer_reparieren(fn)
        else:
            logging.info("Keine doppelten Fallnummer")
            

    def fallnummer_reparieren(self, fn):
        faelle = FallList(where="fn='%s'" % fn, order="bgy,bgm,bgd")
        for f in faelle[1:]:
            try:
                logging.info("Versuche Fallnummer zu reparieren: fn=%(fn)s id=%(id)s  akte_id=%(akte_id)s" % f)
                akte_id = f['akte_id']
                if akte_id:
                    try:
                        akte = f['akte']
                        logging.info("  Akte: na=%(na)s vn=%(vn)s  id=%(id)s" % akte)
                    except Exception, m:
                        logging.error("  Kein Aktenobjekt: %s" % m)
                        raise
                else:
                    logging.error("  Keine Akten-Id")
                    raise Exception("  Keine Akten-Id")
                stz_code = f['akte__stzbg__code']
                jahr = f['bgy']
                new_fn = getNewFallnummer(stz_code, jahr)
                f.update({'fn': new_fn})
                fall_id = f['id'] 
                where = 'fall_id=%s' % fall_id
                for s in chain(JugendhilfestatistikList(where=where),
                               Jugendhilfestatistik2007List(where=where),
                               FachstatistikList(where=where),
                               ):
                    s.update({'fall_fn': new_fn})
                logging.info("Fallnummer repariert: fall_id=%s: %s --> %s" %
                             (fall_id, fn, new_fn))
            except:
                logging.error(
                    "****Fallnummer konnte nicht repariert werden:  "
                    "fn=%(fn)s id=%(id)s  akte_id=%(akte_id)s" % f)
                                 
            
    def indexe_bei_bedarf_einfuehren(self):
        for table, index_name, column in (
            ('fall', 'akte_id', 'akte_id'),
            ('fall', 'bgy', 'bgy'),
            ('fall', 'zday', 'zday'),
            ('bezugsperson', 'akte_id', 'akte_id'),
            ('einrichtung', 'akte_id', 'akte_id'),
            ('leistung', 'fall_id', 'fall_id'),
            ('leistung', 'mit_id', 'mit_id'),
            ('anmeldung', 'fall_id', 'fall_id'),
            ('zustaendigkeit', 'fall_id', 'fall_id'),
            ('zustaendigkeit', 'mit_id', 'mit_id'),
            ('zustaendigkeit', 'bgy', 'bgy'),
            ('zustaendigkeit', 'ey', 'ey'),
            ('dokument', 'fall_id', 'fall_id'),
            ('dokument', 'mit_id', 'mit_id'),
            ('fachstat', 'fall_id', 'fall_id'),
            ('fachstat', 'mit_id', 'mit_id'),
            ('jghstat', 'fall_id', 'fall_id'),
            ('jghstat', 'mit_id', 'mit_id'),
            ('jghstat07', 'fall_id', 'fall_id'),
            ('jghstat07', 'mit_id', 'mit_id'),
            ('wireb', 'fall_id', 'fall_id'),
            ('wireb', 'fallcodenummer', 'fallcodenummer'),
            ('fallberatungskontakt', 'fall_id', 'fall_id'),
            ('fallberatungskontakt', 'bkont_id', 'bkont_id'),
            ('mitarbeiterberatungskontakt', 'mit_id', 'mit_id'),
            ('mitarbeiterberatungskontakt', 'bkont_id', 'bkont_id'),
            ('mitarbeiterfua_bs', 'mit_id', 'mit_id'),
            ('mitarbeiterfua_bs', 'fua_bs_id', 'fua_bs_id'),
            ):
            try:
                SQL("""ALTER TABLE %s ADD INDEX %s (%s)""" % (table, index_name, column)).execute()
                logging.info("Index '%s' fuer Tabelle '%s' angelegt" % (index_name, table))
            except SQLError as e:
                if not 'Duplicate key name' in e[1]:
                    raise
                #logging.info("Index '%s' fuer Tabelle '%s' existiert bereits" % (index_name, table))
    def add_table(self, tabelle, name, klasse):
        """Tabelle anlegen mit nur einem Feld, dem PrimaryKey.
        Weitere Felder mit add_field() hinzufgen.
        """
        if tabelle in self.tables:
            logging.info("Tabelle '%s' existiert bereits" % (name,))
            return
        try:
            SQL("""CREATE TABLE %s (
id int NOT NULL,
PRIMARY KEY (id)
) ENGINE=MyISAM DEFAULT CHARSET=latin1""" % (tabelle,)).execute()
            t = Tabelle()
            t.init(
                tabelle=tabelle,
                name=name,
                klasse=klasse,
                flag=0,
                maxist=0
                )
            t.new()
            t.insert()
            felder = (
                ('id', 'id', None, 'INT', cc('verwtyp', 's'), None,),
                )
            for feld, name, inverse, typ, verwtyp, ftab_id in felder:
                f = Feld()
                f.init(
                    tab_id=t['id'],
                    feld=feld,
                    name=name,
                    inverse=inverse,
                    typ=typ,
                    verwtyp=verwtyp,
                    ftab_id=ftab_id,
                    flag=0,
                    )
                f.new()
                f.insert()
            logging.info("Neue Tabelle '%s' angelegt" % (tabelle,))
        except:
            logging.error("************ Fehler beim Anlegen der Tabelle '%s'" % (tabelle,))
            raise
        self.tables = getDBHandle().listtables()
        assert tabelle in self.tables

    def wireb_bei_bedarf_nachtragen(self):
        # wireb umgestellt von 0,1,2 auf true,false
        wireb_config = register_get('wireb')
        if wireb_config:
            if wireb_config in  ('1', '2'):
                register_set('wireb', 'true')
            else:
                register_set('wireb', 'false')
        if 'wireb' in self.tables:
            logging.info("Tabelle 'wireb' existiert bereits")
        else:
            self.add_table('wireb', "Wir.EB-Integration", "Wireb")
        self.add_field('wireb', 'fall_id', 'Fall_id', 'INT', 'f',
                       'wireb', None, 'fall')
        self.add_field('wireb', 'fallcodenummer', 'Fallcodenummer', 'INT', 'p')
        self.add_date('wireb', 'fcbg', 'Fallcodenummer bei Wir.EB angelegt')
        self.add_date('wireb', 'jghbg', 'Bundesstatistik nach Wir.EB hochgeladen')
        self.add_field('wireb', 'alle_bearbeitet', 'Alle Evaluationsbgen bearbeitet', 
                       'INT', 'k', kat_code='ja_nein')

    def add_or_rename_code(self, kat_code, code, name, sort=1, dok=''):
        try:
            Code(kat_code=kat_code, code=code)
            self.rename_code(kat_code, code, name, sort=sort, dok=dok)
        except DBAppError:
            self.add_code(kat_code, code, name, sort=sort, dok=dok)

    def update_bundesstatistikdefinition(self):
        """
        Updatet werden nur Items, Optionsgruppen, Optionen
        Es werden keine Items gelscht.
        Es knnen keine Felder in der jghstat07-Tabelle hinzugefgt werden.
        Stattdessen mssen die Jokerfelder verwendet werden (4 Optionsfelder,
        2 Ganzzahlfelder, 2 Textfelder).
        Das wird schon bei jghdef.validate() geprft.
        Es wird kein Typ eines Items gendert.
        Es werden keine Optionsgruppen (kat_code) gelscht.
        Die Optionsgruppe (kat_code) fr ein Item darf nicht gendert werden.
        (wird schon bei der jghdef.validate() geprft).
        Es werden keine Optionen gelscht.

        Statistik wird bei Programmstart eingelesen und validiert.
        Dies geschieht in updatebundesstatistik.__init__()
        Wenn keine Statistik fr das aktuelle Jahr da ist:
        - Loggen, aber keine Exception
        Statistik fr knftige Jahre wird nicht eingelesen.
        Statistik fr vergangene Jahr wird eingelesen. (Wieviele?)
        
        # Bis jetzt nicht:
        # Statistik solle EBKuS-Version haben, ab der die Statistik verwendbar ist.
        # (ebkus_version_ab), Fehler, wenn das nicht gegeben ist.

        Beim Einlesen wird das Statistikobjekt validiert. Wenn es nicht valide ist,
        kann EBKuS nicht starten.

        Beim Update kann es Fehler geben. Dann startet EBKuS ebenfalls nicht.


        Die Datenbank wird immer nur auf der Grundlage der neuesten Statistikdefinition 
        updatet, und auch nur dann, wenn die Datei mit der Statistikdefinition sich
        gendert hat, festgestellt am Datum der letzten nderung.
        Nach erfolgreichem Lesen der Bundesstatistik fr ein Jahr wird im Register
        ein Eintrag gemacht: bundesstatistik_<jahr>: <letzte nderung von jghdef_<jahr>.
        Falls eine neuere Datei vorliegt, wird komplettes Update gefahren, sonst nicht.
        Die alten Statistiken werden auch nach diesem Verfahren durchgegangen (alle?
        wie weit zurck?)

        
        """
        # geht erst ab 4.13, aber wenn dieser code luft, sind wir ja schon so weit.
        # jghdef wird hier beim Start des Servers initialisiert.
        ubs = UpdateBundesstatistikdefinition(self)
        if ubs.update_noetig():
            ubs.update()
        
class UpdateBundesstatistikdefinition(object):
    timestamp_format = "%Y%m%d_%H%M%S"
    def __init__(self, u):
        self.u = u
        # die neueste der nicht in der Zukunft liegenden:
        self.jghdef = get_jghdef(strict=False)
        self.jahr = self.jghdef.berichtsjahr
        #print 'NEWEST', self.jahr
    def update_noetig(self):
        bm = get_bundesstatistikmanager()
        path = bm.get_path_for_year(self.jahr)
        #print 'PATH', path
        last_file_change = getmtime_as_dtstr(path)
        last_update_str = register_get('last_update_bundesstatistikdefinition_{}'.format(self.jahr))
        noetig = False
        if last_update_str is None or last_file_change > last_update_str:
            noetig = True
        logging.info("Bundesstatistik-Update {} {}noetig".format(
            self.jghdef.berichtsjahr, '' if noetig else 'nicht '))
        return noetig
    def update(self):
        # do something
        self.jgh_db_felder = dict((f['feld'], f) for f in FeldList(
            where="tab_id = {}".format(Tabelle(tabelle='jghstat07')['id'])))
        for item in self.jghdef.fb_item_iter():
            if item.berechnet or item.fester_wert:
                continue
            self.u.rename_field_langname('jghstat07', item.ebkus_name, item.langname)
            if item.typ == 'Option':
                self.update_option(item)
            elif item.typ == 'Ganzzahl':
                self.update_ganzzahl(item)
            elif item.typ == 'Zeichenkette':
                self.update_zeichenkette(item)
            elif item.typ == 'Checkbox':
                self.update_checkbox(item)
        # logging.info("Update fr Bundesstatistik {} durchgefhrt (dry-run).".format(self.jghdef.berichtsjahr))
        logging.info("Bundesstatistik-Update {} durchgefuehrt".format(self.jghdef.berichtsjahr))
        
        register_set('last_update_bundesstatistikdefinition_{}'.format(self.jahr),
                     dtstr_from_dt(datetime.datetime.now()))
    def update_option(self, item):
        kat_code = item.optionsgruppe
        # Optionen updaten
        # Alle codes mssen in der DB stehen
        # Der Text in code.name muss identisch sein mit option.text
        # Wenn man die Codes aus der Datenbank nimmt, muss die Sortierung,
        # die sich aus code.sort ergibt, mit der Reihenfolge in optionen
        # bereinstimmen.
        # codes, die nicht mehr verwendet werden, werden *nicht* auf ungltig
        # gestellt. Das kann man immer noch manuell tun, dann tauchen sie
        # auch bei der Statistikauszhlung nicht mehr auf.
        # 
        self.update_optionen_und_bereiche(item, item.optionsgruppe,
                                          item.optionen, invalidate_others=False)
    def update_optionen_und_bereiche(self, item, kat_code,
                                     optionen_oder_bereiche, invalidate_others=False):
        item_optionen_dict = dict((o.code, o) for o in optionen_oder_bereiche)
        db_optionen_dict = dict((c['code'], c) for c in get_all_codes(kat_code))
        # Fehlende Codes hinzufgen und Sortierung anpassen
        # Es ist theoretisch mglich, dass nicht mehr verwendete Optionen ein sort-Wert haben,
        # der identisch ist mit einer, die verwendet wird. Das ist aber praktisch
        # irrelevant.
        # Die Deaktivierung von Optionen, die nicht in der jghdef aufgefhrt sind,
        # scheint mit (Stand Tue Feb  6 11:13:11 2024) mir fr Bereichskategorien
        # sinnvoll, fr andere Kategorien weniger.
        # Die Auswahl fr die aktuellen Formulare wird ohnehin ber die jghdef gesteuert,
        # und hngt nicht davon ab, ob ein Code deaktiviert ist oder nicht.
        last_sort = 0
        for option in optionen_oder_bereiche:
            code = option.code
            if code is None:
                # Zwischenberschrift ohne code kann auch dabei sein
                continue
            if code in db_optionen_dict:
                codeobj = db_optionen_dict[code]
                sort = codeobj['sort']
                name = codeobj['name']
                new_sort = sort
                new_name = name
                if sort <= last_sort:
                    new_sort = last_sort + 1
                if option.text != codeobj['name']:
                    new_name = option.text
                if new_sort != sort or new_name != name:
                    self.u.rename_code(kat_code, code, new_name, sort=new_sort)
                self.u.activate_code(kat_code, code, 0)
                last_sort = new_sort
            else:
                new_sort = last_sort + 1
                self.u.add_code(kat_code, code, option.text, sort=new_sort)
                last_sort = new_sort
        if invalidate_others:
            for codeobj in db_optionen_dict.values():
                code = codeobj['code']
                if code not in item_optionen_dict:
                    # deaktivieren ab Beginn des Berichtjahrs
                    self.u.activate_code(kat_code, code, 1, dy=self.jahr, dm=1)
        if item.ebkus_name.startswith('jok') and optionen_oder_bereiche:
            try:
                codeobj = Code(kat_code=kat_code, code='1')
                if codeobj['name'] == 'Dummy':
                    codeobj.delete()
            except:
                pass
        
    def update_ganzzahl(self, item):
        if item.bereichsgruppe:
            # Bereiche, die nicht hier definiert sind, werden alle deaktiviert.
            # Denn nur eine Kategorisierung der Werte einer Zahl ist sinnvoll,
            # und macht auch immer Sinn fr vergangene Versionen der Bundes-
            # statistik.
            self.update_optionen_und_bereiche(
                item, item.bereichsgruppe, item.bereiche, invalidate_others=True)
            
    def update_zeichenkette(self, item):
        self.u.varchar_feld_verlaengern('jghstat07', item.ebkus_name, item.feldlaenge)
        
    def update_checkbox(self, item):
        self.u.varchar_feld_verlaengern('jghstat07', item.ebkus_name, item.feldlaenge)

                         
