#!/usr/bin/python
# coding: utf8


"""
EBKuS-Sicherungsdateien in einem Backupverzeichnis gemäß einer
Aufbewahrungsrichtlinie löschen.

Default-Werte für die Aufbewahrungsrichtlinie:
KEEP_DAYS = 60
KEEP_WEEKS = 26
KEEP_MONTHS = 36
KEEP_YEARS = 30
KEEP_VERSIONS = 50

Autor: Albrecht Schmiedel 
Email: albrecht.schmiedel@ebkus.org
"""
# Aufbewahrungsrichtlinie Default-Werte
KEEP_DAYS = 60
KEEP_WEEKS = 26
KEEP_MONTHS = 36
KEEP_YEARS = 30
KEEP_VERSIONS = 50


import re
import os
import sys
from os.path import basename, join, isdir
import argparse
from datetime import datetime, timedelta, date

def mark(objects, # list of tuples sorted desc: (year, month, week, day, version, instance, id)
         keep,    # how many newest for each unit to keep     
         index,   # 0 year, 1 month, 2 week, 3 day, 4 version
         ):
    """
    Der Kernalgorithmus, um die Aufbewahrungsrichtlinie (retention policy) umzusetzen.
    Wird in PruneEBKuSBackups() auf Tuple wie
    (2020, 12, 50, 9, '4.10.1', 'demo', 'demo_backup_v4.10.1_2020-12-09_10-57-01.zip')
    angewendet. 


    Return set of ids to keep.
    >>> objects = (
    ... (6, '4.10.1', 1),
    ... (5, '4.10.1', 2),
    ... (5, '4.10.1', 3),
    ... (5, '4.10', 4),
    ... (5, '4.10', 5),
    ... (2, '4.10', 6),
    ... (7, '4.10', 7),
    ... (7, '4.10', 8),
    ... (6, '4.10', 9),
    ... (6, '4.10', 9),
    ... (6, '4.10', 9),
    ... )
    >>> # Nur einen aufheben
    >>> mark(objects, 1, 0) == set([1,])
    True
    >>> mark(objects, 4, 0) == set([1,2,6,7])
    True
    >>> mark(objects, 4, 1) == set([1,4])
    True
    >>> mark(objects, 2, 1) == set([1,4])
    True
    >>> mark(objects, 1, 1) == set([1,])
    True
    >>> mark(objects, 6, 0) == set([1,2,6,7,9])
    True
    >>> mark(objects, 600, 0) == set([1,2,6,7,9])
    True
    """
    marked = set()
    past = 0
    last = None
    for obj in objects:
        # id muss das letzte Element des Tupels sein
        id = obj[-1] 
        val = obj[index]
        if past == keep:
            # die neuesten keep Element sind markiert
            break
        if val != last:
            marked.add(id)
            last = val
            past += 1
    return marked

# zum parsen von z.B. 'demo_backup_v4.10.1_2020-12-09_10-57-01.zip'
_re_name = re.compile(r'^(\w+)_backup_v([0-9.]+)_(\d{4})-(\d{2})-(\d{2})_(\d{2})-(\d{2})-(\d{2})\.zip')
def parse(filename):
    """
    EBKuS-Sicherungsdateiname parsen und Werttuple zurückgeben.
    None, wenn nicht als EBKuS-Sicherungsdateiname erkennbar.
    Eine zeitlich absteigend sortierte Liste von Wertetupeln kann dann mit mark()
    ausgewertet werden, um die gemäß der Policy beizubehaltenden Backups
    zu finden.
    >>> filename = 'demo_backup_v4.10.1_2020-12-09_10-57-01.zip'
    >>> res = parse(filename)
    >>> res == (2020, 12, 50, 9, '4.10.1', 'demo', 'demo_backup_v4.10.1_2020-12-09_10-57-01.zip')
    True
    """
    match = _re_name.match(filename)
    if match:
        data = match.groups()
    else:
        # Kein gültiger Name für eine EBKuS-Sicherungsdatei
        return None
    dt = datetime(*(int(x) for x in data[2:]))
    t = (dt.year, dt.month, dt.isocalendar()[1], dt.day, data[1], data[0], filename)
    return t

class PruneEBKuSBackups(object):
    def __init__(self, directory, keep_years=30, keep_months=36,
                 keep_weeks=26, keep_days=60, keep_versions=20):
        self.directory = directory
        self.keep_years = keep_years
        self.keep_months = keep_months
        self.keep_weeks = keep_weeks
        self.keep_days = keep_days
        self.keep_versions = keep_versions
        self._init()
    def _init(self):
        self.objects = {} # dict: filename --> tuple, nur die korrekt geparsten
        self.ordered_objects = {} # dict: instance --> ordered list of tuples descending
        self.instances = set() # set of instances with backups in directory parsed from filename
        self.filenames = set() # set of filenames in directory
        self.filenames_to_keep = set() # selected filenames after applying policy not to be deleted
    @property
    def filenames_to_delete(self):
        return set(self.objects).difference(self.filenames_to_keep)
    @property
    def filenames_to_keep_ordered(self):
        objects_to_keep = [self.objects[f] for f in self.filenames_to_keep]
        objects_to_keep.sort()
        return [x[-1] for x in objects_to_keep]
    def parse_filenames(self):
        """Alle Dateiname in self.filenames, die einen EBKuS-Backup-Dateinamen matchen,
        parsen und als dict: Dateiname --> Tuple ablegen.
        Tuple wie z.B. (2020, 12, 50, 9, '4.10.1', 'demo', 
                       'demo_backup_v4.10.1_2020-12-09_10-57-01.zip')
        Dabei Instanznamen aufsammeln und in self.instances ablegen.
        """
        for f in self.filenames:
            t = parse(f)
            if t:
                self.objects[f] = t
                self.instances.add(t[5])
    def ordered_list_for_each_instance(self):
        for instance in self.instances:
            objects_for_instance = [x for x in self.objects.values() if x[5] == instance]
            objects_for_instance.sort()
            objects_for_instance.reverse()
            self.ordered_objects[instance] = objects_for_instance
    def prune(self):
        """Policy auf self.filenames anwenden.
        Übrig bleiben die in self.filenames_to_keep genannten.
        """
        self.parse_filenames()
        self.ordered_list_for_each_instance()
        for instance in self.instances:
            self.prune_instance(instance)
    def prune_instance(self, instance):
        """Die Funktion mark() für alle Zeiteinheit anwenden.
        Das Ergebnis ist die Vereinigung aller Dateinamen, die mindestens
        einem Kriterium genügen.
        """
        objects = self.ordered_objects[instance]
        self.filenames_to_keep.update(mark(objects, self.keep_years, 0))
        self.filenames_to_keep.update(mark(objects, self.keep_months, 1))
        self.filenames_to_keep.update(mark(objects, self.keep_weeks, 2))
        self.filenames_to_keep.update(mark(objects, self.keep_days, 3))
        self.filenames_to_keep.update(mark(objects, self.keep_versions, 4))
    def prune_dir(self, dry_run=False):
        """self.filenames aus self.directory befüllen und prune() drauf anwenden.
        Dann die nicht ausgewählten Dateien löschen.
        """
        self._init()
        self.filenames.update([basename(f) for f in os.listdir(self.directory)])
        self.prune()
        to_delete = self.filenames_to_delete
        for f in to_delete:
            path = join(self.directory, f)
            if not dry_run:
                os.unlink(path)
        assert len(self.filenames_to_keep) + len(to_delete) == len(self.objects)
        return len(self.filenames_to_keep), len(to_delete)
    def create_mockup_filenames(self, startdate, enddate, version, instance):
        """Filenamen für einen Intervall, eine Version und eine Instanz generieren.
        Dateinamen fuer alle Daten zwischen startdate und endate inklusive.
        >>> peb = PruneEBKuSBackups('.', 30, 24, 12, 40, 10)
        >>> peb.create_mockup_filenames(date(2020,11,4), date(2020,11,4), '4.10', 'demo')
        ['demo_backup_v4.10_2020-11-04_03-11-22.zip']
        >>> peb.create_mockup_filenames(date(2020,11,4), date(2020,11,6), '4.10', 'demo')
        ['demo_backup_v4.10_2020-11-04_03-11-22.zip', 'demo_backup_v4.10_2020-11-05_03-11-22.zip', 'demo_backup_v4.10_2020-11-06_03-11-22.zip']
        """
        # filename = 'demo_backup_v4.10.1_2020-12-09_10-57-01.zip'
        filename_format = "{}_backup_v{}_{:04}-{:02}-{:02}_{:02}-{:02}-{:02}.zip"
        filenames = []
        start = (startdate.year, startdate.month, startdate.day)
        end = (enddate.year, enddate.month, enddate.day)
        assert start <= end
        daydate = startdate
        while True:
            year, month, week, day = daydate.year, daydate.month, daydate.isocalendar()[1], daydate.day
            if (year, month, day) > end:
                break
            filename = filename_format.format(instance, version, year, month, day, 3,11,22)
            filenames.append(filename)
            daydate = daydate + timedelta(days=1)
        return filenames
    def create_mockup(self):
        """
        >>> peb = PruneEBKuSBackups('.', 30, 24, 12, 40, 10)
        >>> peb.create_mockup()
        9000
        """
        mockup_data = (
            (date(2010,5,22), date(2011,9,5), '3.4', 'demo'),
            (date(2010,3,22), date(2011,9,5), '3.4', 'demo_berlin'),

            (date(2011,9,6), date(2013,9,5), '3.8', 'demo'),
            (date(2011,9,6), date(2013,9,5), '3.8', 'demo_berlin'),

            (date(2013,9,6), date(2016,9,5), '4.1.3', 'demo'),
            (date(2013,9,6), date(2016,6,5), '4.1.1', 'demo_berlin'),
            (date(2016,6,6), date(2016,9,5), '4.1.3', 'demo_berlin'),

            (date(2016,9,6), date(2018,8,13), '4.8', 'demo'),
            (date(2016,9,6), date(2018,8,13), '4.8', 'demo_berlin'),
            (date(2017,9,6), date(2018,10,13), '4.8', 'demo_braunschweig'),

            (date(2018,8,14), date(2020,10,13), '4.9', 'demo'),
            (date(2018,8,14), date(2020,10,13), '4.9', 'demo_berlin'),
            (date(2018,10,14), date(2020,10,13), '4.9', 'demo_braunschweig'),

            (date(2020,10,14), date(2020,12,14), '4.11.4', 'demo'),
            (date(2020,10,14), date(2020,12,22), '4.11.4', 'demo_berlin'),
            (date(2020,10,14), date(2020,12,29), '4.11.4', 'demo_braunschweig'),
            )
        for t in mockup_data:
            self.filenames.update(self.create_mockup_filenames(*t))
        return len(self.filenames)
    def create_mockup_dir(self):
        """Das Verzeichnis mit Mockup-Dateinamen füllen.
        """
        self._init()
        count = self.create_mockup()
        if not isdir(self.directory):
            os.mkdir(self.directory)
        for f in self.filenames:
            filename = join(self.directory, f)
            open(filename, 'ab')
        return count

def create_args_parser():
    parser = argparse.ArgumentParser(
        description="EBKuS-Backups gemäß einer Aufbewahrungsrichtlinie (retention policy) löschen",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        'directory',
        action='store',
        help='Backup-Verzeichnis, aus dem verzichtbare Backups gelöscht werden sollen.',
    )
    parser.add_argument(
        '--keep-years',
        default=KEEP_YEARS,
        type=int,
        action='store',
        help='Anzahl der Jahre, für die das jeweils neueste Backup aufbewahrt wird.',
    )
    parser.add_argument(
        '--keep-months',
        default=KEEP_MONTHS,
        type=int,
        action='store',
        help='Anzahl der Monate, für die das jeweils neueste Backup aufbewahrt wird.',
    )
    parser.add_argument(
        '--keep-weeks',
        default=KEEP_WEEKS,
        type=int,
        action='store',
        help='Anzahl der Wochen, für die das jeweils neueste Backup aufbewahrt wird.',
    )
    parser.add_argument(
        '--keep-days',
        default=KEEP_DAYS,
        type=int,
        action='store',
        help='Anzahl der Tage, für die das jeweils neueste Backup aufbewahrt wird.',
    )
    parser.add_argument(
        '--keep-versions',
        default=KEEP_VERSIONS,
        type=int,
        action='store',
        help='Anzahl der Versionen, für die das jeweils neueste Backup aufbewahrt wird.',
    )
    parser.add_argument(
        '-n', '--dry-run',
        action='store_true',
        help="Nicht wirklich löschen, nur die zu löschenden Dateien anzeigen."
    )
    parser.add_argument(
        '--create-mockup',
        action='store_true',
        help="Verzeichnis mit Dateinamen generieren ohne echte Daten. Zum Testen."
    )
    parser.add_argument("--verbosity", "-v", action='store_true', 
                        help="Geschwätzigkeit.")
    return parser

def main(cmd_args):
    parser = create_args_parser()
    args = parser.parse_args(cmd_args)
    verbosity =  args.verbosity
    if verbosity:
        # print 'Args: ', args
        pass
    peb = PruneEBKuSBackups(args.directory,
                            keep_years=args.keep_years,
                            keep_months=args.keep_months,
                            keep_weeks=args.keep_weeks,
                            keep_days=args.keep_days,
                            keep_versions=args.keep_versions)
    if args.create_mockup:
        print 'Mockup'
        if 'test' not in args.directory.lower():
            print 'Mockup nur mit Verzeichnisnamen, die die Zeichnkette "test" enthalten'
            sys.exit(1)
        count = peb.create_mockup_dir()
        print 'generiert:', count
        sys.exit(0)
    kept, deleted = peb.prune_dir(dry_run=args.dry_run)
    if verbosity:
        for f in peb.filenames_to_delete:
            print 'geloescht: ', f
    print 'aufbewahrt:', kept, ' geloescht:', deleted


if __name__ == "__main__":
    # import doctest
    # doctest.testmod()
    main(sys.argv[1:])
